! Description:
!> @file
!!   Subroutines for interface to HT-FRTC
!
!> @brief
!!   Subroutines for interface to HT-FRTC
!!
!! @details
!!   The htfrtc_interface subroutine implements the HT-FRTC forward and
!!   Jacobian models. The Jacobian model is run if the optional profiles_k_pc
!!   argument is present.
!
! Copyright:
!    This software was developed within the context of
!    the EUMETSAT Satellite Application Facility on
!    Numerical Weather Prediction (NWP SAF), under the
!    Cooperation Agreement dated 25 November 1998, between
!    EUMETSAT and the Met Office, UK, by one or more partners
!    within the NWP SAF. The partners in the NWP SAF are
!    the Met Office, ECMWF, KNMI and MeteoFrance.
!
!    Copyright 2018, EUMETSAT, All Rights Reserved.
!
MODULE rttov_htfrtc_interface_mod

  USE parkind1, ONLY: jpim, jprb, jplm

  IMPLICIT NONE

  PRIVATE
  PUBLIC :: htfrtc_interface

CONTAINS

  SUBROUTINE htfrtc_interface(err, &
                              coefs, &
                              opts, &
                              profiles, &
                              pccomp, &
                              profiles_k_pc, &
                              profiles_k_rec)
#include "throw.h"
    USE rttov_types, ONLY : &
        rttov_coefs, &
        rttov_options, &
        rttov_profile, &
        rttov_pccomp
    USE rttov_const, ONLY : &
        deg2rad, gravity, earthradius, mair, rgc, &
        planck_c1, planck_c2, &
        surftype_land, surftype_sea, surftype_seaice

    IMPLICIT NONE

! input/output
    INTEGER(jpim)           , INTENT(OUT)              :: err
    TYPE(rttov_coefs)       , INTENT(IN)               :: coefs
    TYPE(rttov_options)     , INTENT(IN)               :: opts
    TYPE(rttov_profile)     , INTENT(IN)               :: profiles(:)
    TYPE(rttov_pccomp)      , INTENT(INOUT)            :: pccomp
    TYPE(rttov_profile)     , INTENT(INOUT) , OPTIONAL :: profiles_k_pc(:)
    TYPE(rttov_profile)     , INTENT(INOUT) , OPTIONAL :: profiles_k_rec(:)

    LOGICAL(jplm) :: do_k

    REAL(jprb),PARAMETER :: zero=0.0
    REAL(jprb),PARAMETER :: unity=1.0
    REAL(jprb),PARAMETER :: two=2.0
    REAL(jprb),PARAMETER :: ahalf=0.5
    REAL(jprb),PARAMETER :: ahundred=100.0
    REAL(jprb),PARAMETER :: ahundredth=0.01

    INTEGER(jpim) :: n_p,n_f,n_t,n_q,n_b,n_lt,n_prof
    INTEGER(jpim) :: nlayers,nlevels,n_pc,n_ch
    INTEGER(jpim) :: surf_level
    INTEGER(jpim) :: i,i_f,i_p,i_pc,i_ch
    INTEGER(jpim) :: ja,jb,js
    INTEGER(jpim) :: t_b_surf_i

    INTEGER(jpim),DIMENSION(profiles(1)%nlayers) :: p_opt_ly_i,t_opt_ly_i,q_opt_ly_i
    INTEGER(jpim),DIMENSION(profiles(1)%nlevels) :: t_b_i
    INTEGER(jpim),DIMENSION(profiles(1)%nlayers,2) :: qt

    REAL(jprb),DIMENSION(profiles(1)%nlevels) :: z
    REAL(jprb),DIMENSION(profiles(1)%nlevels) :: t_b_r,b
    REAL(jprb),DIMENSION(profiles(1)%nlevels) :: rdown,rup
    REAL(jprb),DIMENSION(profiles(1)%nlevels) :: dbdt
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: p_ly,dp,dz
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: t_ly,t2_ly
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: p_opt_ly_r,t_opt_ly_r,q_opt_ly_r
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: blydown,blyup,lt
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: q_ly,ln_q_ly,o3_ly
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: tausl,tr
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: gravm1,mpath,mft
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: dblydowndt,dblyupdt
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: totdown,totup
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: dq,do3,dsg
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: dtrdt,dtrdq,dtrdo3
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: drdowndt,drdowndq,drdowndo3
    REAL(jprb),DIMENSION(profiles(1)%nlayers) :: drupdt,drupdq,drupdo3
    REAL(jprb),DIMENSION(profiles(1)%nlayers,4) :: pt,pq
    REAL(jprb),DIMENSION(profiles(1)%nlayers,8) :: ptq
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_f) :: rc,normm1
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_f,profiles(1)%nlevels) :: drcdt,drcdq,drcdo3
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_f) :: drcdst
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_f) :: drcds2mt,drcds2mq
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_f) :: drcds2mp,drcds2mu,drcds2mv
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_f) :: surf_em,aems,bems,cems,expf
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_f) :: daems,dbems,dcems,dexpf,dsurf_em
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_f) :: dsurf_emdu,dsurf_emdv
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_ch) :: wn,wn3
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_ch) :: planck_c1_wn3,planck_c2_wn
    REAL(jprb),DIMENSION(coefs%coef_htfrtc%n_ch) :: swrad_tmp

    REAL(jprb) :: t_b_min,t_b_inc,t_b_inc_m1
    REAL(jprb) :: lt_min,lt_inc,lt_inc_m1
    REAL(jprb) :: t_opt_min,t_opt_inc,t_opt_inc_m1
    REAL(jprb) :: q_opt_min,q_opt_inc,q_opt_inc_m1
    REAL(jprb) :: surf_wind,surf_wind2,surf_windm1
    REAL(jprb) :: surf_layer_frac,t_b_surf_r
    REAL(jprb) :: rdown_surf,b_surf
    REAL(jprb) :: db_surfdt
    REAL(jprb) :: tot
    REAL(jprb) :: sg_ly

    TRY

    IF (opts%htfrtc_opts%reconstruct .AND. PRESENT(profiles_k_rec) .AND. &
        .NOT. PRESENT(profiles_k_pc)) THEN
      err = errorstatus_fatal
      THROWM(err.NE.0,'HTFRTC K model requires both profiles_k_pc and profiles_k_rec for rec. rads')
    ENDIF

    sg_ly=coefs%coef_htfrtc%val_mf_l(coefs%coef_htfrtc%n_gas_l)
    n_ch=coefs%coef_htfrtc%n_ch
    nlevels=profiles(1)%nlevels
    nlayers=profiles(1)%nlayers
    n_f=coefs%coef_htfrtc%n_f
    n_p=coefs%coef_htfrtc%n_p
    n_b=coefs%coef_htfrtc%n_b
    n_lt=coefs%coef_htfrtc%n_lt
    n_prof=size(profiles)
    do_k=PRESENT(profiles_k_pc)

    IF(coefs%coef_htfrtc%opt_prop_type==1) THEN
       t_opt_min=coefs%coef_htfrtc%val_l(1)
       t_opt_inc=coefs%coef_htfrtc%val_l(2)-coefs%coef_htfrtc%val_l(1)
       t_opt_inc_m1=unity/t_opt_inc
       n_t=coefs%coef_htfrtc%n_val_l
       q_opt_min=log(coefs%coef_htfrtc%val_mf_nl(1,1))
       q_opt_inc=log(coefs%coef_htfrtc%val_mf_nl(2,1))-log(coefs%coef_htfrtc%val_mf_nl(1,1))
       q_opt_inc_m1=unity/q_opt_inc
       n_q=coefs%coef_htfrtc%n_mf_nl
    ENDIF

    t_b_min=coefs%coef_htfrtc%val_b(1)
    t_b_inc=coefs%coef_htfrtc%val_b(2)-coefs%coef_htfrtc%val_b(1)
    t_b_inc_m1=unity/t_b_inc

    lt_min=coefs%coef_htfrtc%val_lt(1)
    lt_inc=coefs%coef_htfrtc%val_lt(2)-coefs%coef_htfrtc%val_lt(1)
    lt_inc_m1=unity/lt_inc

    DO i_f=1,n_f
       normm1(i_f)=1.0/coefs%coef_htfrtc%val_norm(i_f)
    ENDDO

    n_pc=coefs%coef_htfrtc%n_pc
    IF((opts%htfrtc_opts%n_pc_in.GT.0).AND.(opts%htfrtc_opts%n_pc_in.LE.coefs%coef_htfrtc%n_pc)) THEN
       n_pc=opts%htfrtc_opts%n_pc_in
    ENDIF

    IF(opts%htfrtc_opts%reconstruct) THEN
      FORALL(i=1:n_ch)
        wn(i)=coefs%coef_htfrtc%sensor_freq(i)
        wn3(i)=wn(i)**3
        planck_c1_wn3(i)=planck_c1*wn3(i)
        planck_c2_wn(i)=planck_c2*wn(i)
      ENDFORALL
    ENDIF

    !Loop over profiles
    DO i_p=1,n_prof

       !Non-directional horizontal wind
       IF(profiles(i_p)%skin%surftype==surftype_sea) THEN
          surf_wind=sqrt(profiles(i_p)%s2m%u**2 + profiles(i_p)%s2m%v**2)
          surf_wind2=surf_wind**2
          surf_windm1=1.0/(surf_wind+1.0e-6)
       ENDIF

       !Find the level just above the surface
       surf_level=nlevels-1
       surf_layer_frac=1.0
       DO jb=1,nlevels-1
          IF ((profiles(i_p)%s2m%p.GT.profiles(i_p)%p(jb)).AND.(profiles(i_p)%s2m%p.LE.profiles(i_p)%p(jb+1))) THEN
             surf_level=jb
             surf_layer_frac=(profiles(i_p)%s2m%p-profiles(i_p)%p(jb))/(profiles(i_p)%p(jb+1)-profiles(i_p)%p(jb))
          ENDIF
       ENDDO

       FORALL(i=1:nlevels)
          t_b_i(i)=min(n_b-1,max(1,1+int(t_b_inc_m1*(profiles(i_p)%t(i)-t_b_min))))
          t_b_r(i)=t_b_inc_m1*((profiles(i_p)%t(i)-t_b_min)-t_b_inc*(t_b_i(i)-1))
       ENDFORALL
       t_b_i(surf_level+1)=min(n_b-1,max(1,1+int(t_b_inc_m1*(profiles(i_p)%s2m%t-t_b_min))))
       t_b_r(surf_level+1)=t_b_inc_m1*((profiles(i_p)%s2m%t-t_b_min)-t_b_inc*(t_b_i(surf_level+1)-1))
       t_b_surf_i=min(n_b-1,max(1,1+int(t_b_inc_m1*(profiles(i_p)%skin%t-t_b_min))))
       t_b_surf_r=t_b_inc_m1*((profiles(i_p)%skin%t-t_b_min)-t_b_inc*(t_b_surf_i-1))

       !User levels to layers
       FORALL(i=1:nlayers)
          p_ly(i)=ahalf*(profiles(i_p)%p(i)+profiles(i_p)%p(i+1))
          dp(i)=profiles(i_p)%p(i+1)-profiles(i_p)%p(i)
          gravm1(i)=unity/gravity
          mpath(i)=ahundred*dp(i)*gravm1(i)
          t_ly(i)=ahalf*(profiles(i_p)%t(i)+profiles(i_p)%t(i+1))
          q_ly(i)=ahalf*(profiles(i_p)%q(i)+profiles(i_p)%q(i+1))
          o3_ly(i)=ahalf*(profiles(i_p)%o3(i)+profiles(i_p)%o3(i+1))
          dz(i)=dp(i)*gravm1(i)*rgc*t_ly(i)/mair/p_ly(i)
          mft(i)=unity/cos(profiles(i_p)%zenangle*deg2rad)
       ENDFORALL

       !Near surface layer
       p_ly(surf_level)=ahalf*(profiles(i_p)%p(surf_level)+profiles(i_p)%s2m%p)
       dp(surf_level)=profiles(i_p)%s2m%p-profiles(i_p)%p(surf_level)
       gravm1(surf_level)=unity/gravity
       mpath(surf_level)=ahundred*dp(surf_level)*gravm1(surf_level)
       t_ly(surf_level)=ahalf*(profiles(i_p)%t(surf_level)+profiles(i_p)%s2m%t)
       IF(opts%rt_all%use_q2m) THEN
         q_ly(surf_level)=ahalf*(profiles(i_p)%q(surf_level)+profiles(i_p)%s2m%q)
       ELSE
         q_ly(surf_level)=profiles(i_p)%q(surf_level)+surf_layer_frac*(profiles(i_p)%q(surf_level+1)-profiles(i_p)%q(surf_level))
       ENDIF
       o3_ly(surf_level)=profiles(i_p)%o3(surf_level)+surf_layer_frac*(profiles(i_p)%o3(surf_level+1)-profiles(i_p)%o3(surf_level))
       dz(surf_level)=dp(surf_level)*gravm1(surf_level)*rgc*t_ly(surf_level)/mair/p_ly(surf_level)

       z(surf_level+1)=profiles(i_p)%elevation
       DO i=surf_level,1,-1
          z(i)=z(i+1)+dz(i)
       ENDDO

       FORALL(i=1:surf_level)
          mft(i)=unity/sqrt(unity-(sin(profiles(i_p)%zenangle*deg2rad)*earthradius/(earthradius+ahalf*(z(i)+z(i+1))))**2)
       ENDFORALL

       js=1
       DO jb=1,nlayers
          IF (p_ly(jb)<coefs%coef_htfrtc%p(1)) THEN
             p_opt_ly_i(jb)=1
             p_opt_ly_r(jb)=zero
          ELSE IF (p_ly(jb)>=coefs%coef_htfrtc%p(n_p)) THEN
             p_opt_ly_i(jb)=n_p-1
             p_opt_ly_r(jb)=unity
          ELSE
          DO ja=js,n_p-1
             IF (p_ly(jb)>=coefs%coef_htfrtc%p(ja) .AND. p_ly(jb)<coefs%coef_htfrtc%p(ja+1)) THEN
                js=ja
                p_opt_ly_i(jb)=ja
                p_opt_ly_r(jb)=(p_ly(jb)-coefs%coef_htfrtc%p(ja))/(coefs%coef_htfrtc%p(ja+1)-coefs%coef_htfrtc%p(ja))
             ENDIF
          ENDDO
          ENDIF
       ENDDO

       IF(coefs%coef_htfrtc%opt_prop_type==1) THEN
       FORALL(i=1:nlayers)

          t_opt_ly_i(i)=min(n_t-1,max(1,1+int(t_opt_inc_m1*(t_ly(i)-t_opt_min))))
          t_opt_ly_r(i)=t_opt_inc_m1*((t_ly(i)-t_opt_min)-t_opt_inc*(t_opt_ly_i(i)-1))

          ln_q_ly(i)=log(q_ly(i))
          q_opt_ly_i(i)=min(n_q-1,max(1,1+int(q_opt_inc_m1*(ln_q_ly(i)-q_opt_min))))
          q_opt_ly_r(i)=(q_ly(i)-coefs%coef_htfrtc%val_mf_nl(q_opt_ly_i(i),1)) &
          /(coefs%coef_htfrtc%val_mf_nl(q_opt_ly_i(i)+1,1)-coefs%coef_htfrtc%val_mf_nl(q_opt_ly_i(i),1))

          pt(i,1)=(1.0-p_opt_ly_r(i))*(1.0-t_opt_ly_r(i))
          pt(i,2)=(1.0-p_opt_ly_r(i))*t_opt_ly_r(i)
          pt(i,3)=p_opt_ly_r(i)      *(1.0-t_opt_ly_r(i))
          pt(i,4)=p_opt_ly_r(i)      *t_opt_ly_r(i)

          pq(i,1)=(1.0-p_opt_ly_r(i))*(1.0-q_opt_ly_r(i))
          pq(i,2)=(1.0-p_opt_ly_r(i))*q_opt_ly_r(i)
          pq(i,3)=p_opt_ly_r(i)      *(1.0-q_opt_ly_r(i))
          pq(i,4)=p_opt_ly_r(i)      *q_opt_ly_r(i)

          ptq(i,1)=pt(i,1)*(1.0-q_opt_ly_r(i))
          ptq(i,2)=pt(i,1)*q_opt_ly_r(i)
          ptq(i,3)=pt(i,2)*(1.0-q_opt_ly_r(i))
          ptq(i,4)=pt(i,2)*q_opt_ly_r(i)
          ptq(i,5)=pt(i,3)*(1.0-q_opt_ly_r(i))
          ptq(i,6)=pt(i,3)*q_opt_ly_r(i)
          ptq(i,7)=pt(i,4)*(1.0-q_opt_ly_r(i))
          ptq(i,8)=pt(i,4)*q_opt_ly_r(i)

          qt(i,1)=(t_opt_ly_i(i)-1)*n_q+q_opt_ly_i(i)
          qt(i,2)=t_opt_ly_i(i)    *n_q+q_opt_ly_i(i)

       ENDFORALL
       ELSE IF(coefs%coef_htfrtc%opt_prop_type==2) THEN
       FORALL(i=1:nlayers)
          t2_ly(i)=t_ly(i)*t_ly(i)
       ENDFORALL
       ENDIF

          !Loop over frequencies
          IF(profiles(i_p)%skin%surftype==surftype_sea) THEN
             FORALL(i_f=1:n_f)
             aems(i_f)=coefs%coef_htfrtc%coef_ssemp(1,i_f) &
                +surf_wind*coefs%coef_htfrtc%coef_ssemp(2,i_f) &
                +surf_wind2*coefs%coef_htfrtc%coef_ssemp(3,i_f)
             bems(i_f)=coefs%coef_htfrtc%coef_ssemp(4,i_f) &
                +surf_wind*coefs%coef_htfrtc%coef_ssemp(5,i_f) &
                +surf_wind2*coefs%coef_htfrtc%coef_ssemp(6,i_f)
             cems(i_f)=coefs%coef_htfrtc%coef_ssemp(7,i_f) &
                +surf_wind*coefs%coef_htfrtc%coef_ssemp(8,i_f)
             expf(i_f)=exp( ( (coefs%coef_htfrtc%coef_ssemp(9,i_f)-60.0)**2 &
                - (profiles(i_p)%zenangle-coefs%coef_htfrtc%coef_ssemp(9,i_f))**2 ) /cems(i_f) )
             surf_em(i_f)=aems(i_f)+(bems(i_f)-aems(i_f))*expf(i_f)
             ENDFORALL
          ELSE IF (profiles(i_p)%skin%surftype==surftype_land) THEN
             surf_em=0.98
          ELSE IF (profiles(i_p)%skin%surftype==surftype_seaice) THEN
             surf_em=0.99
          ELSE
          ENDIF

          IF(do_k) THEN

             IF(profiles(i_p)%skin%surftype==surftype_sea) THEN
                FORALL(i_f=1:n_f)
                daems(i_f)=surf_windm1*(coefs%coef_htfrtc%coef_ssemp(2,i_f) &
                          +2.0*surf_wind*coefs%coef_htfrtc%coef_ssemp(3,i_f))
                dbems(i_f)=surf_windm1*(coefs%coef_htfrtc%coef_ssemp(5,i_f) &
                          +2.0*surf_wind*coefs%coef_htfrtc%coef_ssemp(6,i_f))
                dcems(i_f)=surf_windm1*coefs%coef_htfrtc%coef_ssemp(8,i_f)
                dexpf(i_f)= -( (coefs%coef_htfrtc%coef_ssemp(9,i_f)-60.0)**2 &
                - (profiles(i_p)%zenangle-coefs%coef_htfrtc%coef_ssemp(9,i_f))**2 ) &
                / (cems(i_f)*cems(i_f))
                dsurf_em(i_f)=(daems(i_f)+(dbems(i_f)-daems(i_f)) &
                             +dcems(i_f)*(bems(i_f)-aems(i_f))*dexpf(i_f))*expf(i_f)
                dsurf_emdu(i_f)=dsurf_em(i_f)*profiles(i_p)%s2m%u
                dsurf_emdv(i_f)=dsurf_em(i_f)*profiles(i_p)%s2m%v
                ENDFORALL
             ELSE
                dsurf_emdu=0.0
                dsurf_emdv=0.0
             ENDIF

          ENDIF

          DO i_f=1,n_f

             IF(coefs%coef_htfrtc%opt_prop_type==1) THEN

                FORALL(i=1:nlayers)
                dq(i)=ptq(i,1)*coefs%coef_htfrtc%coef_nl(qt(i,1)  , p_opt_ly_i(i),  1,i_f) &
                     +ptq(i,2)*coefs%coef_htfrtc%coef_nl(qt(i,1)+1, p_opt_ly_i(i),  1,i_f) &
                     +ptq(i,3)*coefs%coef_htfrtc%coef_nl(qt(i,2)  , p_opt_ly_i(i),  1,i_f) &
                     +ptq(i,4)*coefs%coef_htfrtc%coef_nl(qt(i,2)+1, p_opt_ly_i(i),  1,i_f) &
                     +ptq(i,5)*coefs%coef_htfrtc%coef_nl(qt(i,1)  , p_opt_ly_i(i)+1,1,i_f) &
                     +ptq(i,6)*coefs%coef_htfrtc%coef_nl(qt(i,1)+1, p_opt_ly_i(i)+1,1,i_f) &
                     +ptq(i,7)*coefs%coef_htfrtc%coef_nl(qt(i,2)  , p_opt_ly_i(i)+1,1,i_f) &
                     +ptq(i,8)*coefs%coef_htfrtc%coef_nl(qt(i,2)+1, p_opt_ly_i(i)+1,1,i_f)
                do3(i)=pt(i,1)*coefs%coef_htfrtc%coef_l(t_opt_ly_i(i),  p_opt_ly_i(i),  1,i_f) &
                      +pt(i,2)*coefs%coef_htfrtc%coef_l(t_opt_ly_i(i)+1,p_opt_ly_i(i),  1,i_f) &
                      +pt(i,3)*coefs%coef_htfrtc%coef_l(t_opt_ly_i(i),  p_opt_ly_i(i)+1,1,i_f) &
                      +pt(i,4)*coefs%coef_htfrtc%coef_l(t_opt_ly_i(i)+1,p_opt_ly_i(i)+1,1,i_f)
                dsg(i)=pt(i,1)*coefs%coef_htfrtc%coef_l(t_opt_ly_i(i),  p_opt_ly_i(i),  2,i_f) &
                      +pt(i,2)*coefs%coef_htfrtc%coef_l(t_opt_ly_i(i)+1,p_opt_ly_i(i),  2,i_f) &
                      +pt(i,3)*coefs%coef_htfrtc%coef_l(t_opt_ly_i(i),  p_opt_ly_i(i)+1,2,i_f) &
                      +pt(i,4)*coefs%coef_htfrtc%coef_l(t_opt_ly_i(i)+1,p_opt_ly_i(i)+1,2,i_f)
                ENDFORALL

             ELSE IF (coefs%coef_htfrtc%opt_prop_type==2) THEN

                FORALL(i=1:nlayers)
                dq(i)=((1.0-p_opt_ly_r(i)) *(coefs%coef_htfrtc%coef_nl(1,p_opt_ly_i(i)  ,1,i_f)  &
                                   +t_ly(i) *coefs%coef_htfrtc%coef_nl(2,p_opt_ly_i(i)  ,1,i_f)  &
                                   +t2_ly(i)*coefs%coef_htfrtc%coef_nl(3,p_opt_ly_i(i)  ,1,i_f)  &
                                   +         q_ly(i)*coefs%coef_htfrtc%coef_nl(4,p_opt_ly_i(i)  ,1,i_f)  &
                                   +t_ly(i) *q_ly(i)*coefs%coef_htfrtc%coef_nl(5,p_opt_ly_i(i)  ,1,i_f)  &
                                   +t2_ly(i)*q_ly(i)*coefs%coef_htfrtc%coef_nl(6,p_opt_ly_i(i)  ,1,i_f)) &
                      +p_opt_ly_r(i)       *(coefs%coef_htfrtc%coef_nl(1,p_opt_ly_i(i)+1,1,i_f)  &
                                   +t_ly(i) *coefs%coef_htfrtc%coef_nl(2,p_opt_ly_i(i)+1,1,i_f)  &
                                   +t2_ly(i)*coefs%coef_htfrtc%coef_nl(3,p_opt_ly_i(i)+1,1,i_f)  &
                                   +         q_ly(i)*coefs%coef_htfrtc%coef_nl(4,p_opt_ly_i(i)+1,1,i_f)  &
                                   +t_ly(i) *q_ly(i)*coefs%coef_htfrtc%coef_nl(5,p_opt_ly_i(i)+1,1,i_f)  &
                                   +t2_ly(i)*q_ly(i)*coefs%coef_htfrtc%coef_nl(6,p_opt_ly_i(i)+1,1,i_f)))
                do3(i)=((1.0-p_opt_ly_r(i))*(coefs%coef_htfrtc%coef_l(1,p_opt_ly_i(i)  ,1,i_f)  &
                                  +t_ly(i) *coefs%coef_htfrtc%coef_l(2,p_opt_ly_i(i)  ,1,i_f)  &
                                  +t2_ly(i)*coefs%coef_htfrtc%coef_l(3,p_opt_ly_i(i)  ,1,i_f)) &
                       +p_opt_ly_r(i)      *(coefs%coef_htfrtc%coef_l(1,p_opt_ly_i(i)+1,1,i_f)  &
                                  +t_ly(i) *coefs%coef_htfrtc%coef_l(2,p_opt_ly_i(i)+1,1,i_f)  &
                                  +t2_ly(i)*coefs%coef_htfrtc%coef_l(3,p_opt_ly_i(i)+1,1,i_f)))
                dsg(i)=((1.0-p_opt_ly_r(i))*(coefs%coef_htfrtc%coef_l(1,p_opt_ly_i(i)  ,2,i_f)  &
                                  +t_ly(i) *coefs%coef_htfrtc%coef_l(2,p_opt_ly_i(i)  ,2,i_f)  &
                                  +t2_ly(i)*coefs%coef_htfrtc%coef_l(3,p_opt_ly_i(i)  ,2,i_f)) &
                       +p_opt_ly_r(i)      *(coefs%coef_htfrtc%coef_l(1,p_opt_ly_i(i)+1,2,i_f)  &
                                  +t_ly(i) *coefs%coef_htfrtc%coef_l(2,p_opt_ly_i(i)+1,2,i_f)  &
                                  +t2_ly(i)*coefs%coef_htfrtc%coef_l(3,p_opt_ly_i(i)+1,2,i_f)))
                ENDFORALL

             ENDIF

             FORALL(i=1:nlayers)
                tausl(i)=mpath(i)*mft(i)*(q_ly(i)*dq(i)+o3_ly(i)*do3(i)+sg_ly*dsg(i))
             ENDFORALL

             tr=exp(-tausl)

             FORALL(i=1:nlevels)
                b(i)=coefs%coef_htfrtc%coef_b(t_b_i(i),i_f) &
                    +t_b_r(i)*(coefs%coef_htfrtc%coef_b(t_b_i(i)+1,i_f)-coefs%coef_htfrtc%coef_b(t_b_i(i),i_f))
             ENDFORALL
             b_surf=coefs%coef_htfrtc%coef_b(t_b_surf_i,i_f) &
                    +t_b_surf_r*(coefs%coef_htfrtc%coef_b(t_b_surf_i+1,i_f)-coefs%coef_htfrtc%coef_b(t_b_surf_i,i_f))

             FORALL(i=1:nlayers)
                lt(i)=coefs%coef_htfrtc%coef_lt(min(n_lt,max(1,1+int(lt_inc_m1*(tausl(i)-lt_min)))))
                blydown(i)=b(i)+lt(i)*(b(i+1)-b(i))
                blyup(i)=b(i+1)+lt(i)*(b(i)-b(i+1))
             ENDFORALL

             !Radiative transfer, downwelling
             rdown(1)=0.0
             DO i=2,nlevels
                rdown(i)=blydown(i-1)+(rdown(i-1)-blydown(i-1))*tr(i-1)
             ENDDO

             rdown_surf=blydown(surf_level)+(rdown(surf_level)-blydown(surf_level))*tr(surf_level)

             rup(surf_level+1)=b_surf*surf_em(i_f)+rdown_surf*(unity-surf_em(i_f))
             rup(surf_level)=blyup(surf_level)+(rup(surf_level+1)-blyup(surf_level))*tr(surf_level)
             DO i=surf_level,2,-1
                rup(i-1)=blyup(i-1)+(rup(i)-blyup(i-1))*tr(i-1)
             ENDDO

             rc(i_f)=(rup(1)-coefs%coef_htfrtc%val_mean(i_f))*normm1(i_f)

             IF(do_k) THEN

             IF(coefs%coef_htfrtc%opt_prop_type==1) THEN

               FORALL(i=1:nlayers)
                  dtrdt(i)=-tr(i)*mpath(i)*mft(i)*t_opt_inc_m1* &
                          (q_ly(i)*(pq(i,1)*(coefs%coef_htfrtc%coef_nl(qt(i,2)  , p_opt_ly_i(i),  1,i_f) &
                                            -coefs%coef_htfrtc%coef_nl(qt(i,1)  , p_opt_ly_i(i),  1,i_f)) &
                                   +pq(i,2)*(coefs%coef_htfrtc%coef_nl(qt(i,2)+1, p_opt_ly_i(i),  1,i_f) &
                                            -coefs%coef_htfrtc%coef_nl(qt(i,1)+1, p_opt_ly_i(i),  1,i_f)) &
                                   +pq(i,3)*(coefs%coef_htfrtc%coef_nl(qt(i,2)  , p_opt_ly_i(i)+1,1,i_f) &
                                            -coefs%coef_htfrtc%coef_nl(qt(i,1)  , p_opt_ly_i(i)+1,1,i_f)) &
                                   +pq(i,4)*(coefs%coef_htfrtc%coef_nl(qt(i,2)+1, p_opt_ly_i(i)+1,1,i_f) &
                                            -coefs%coef_htfrtc%coef_nl(qt(i,1)+1, p_opt_ly_i(i)+1,1,i_f))) &
                          +o3_ly(i)*((1.0-p_opt_ly_r(i))*(-coefs%coef_htfrtc%coef_l(t_opt_ly_i(i),  p_opt_ly_i(i),  1,i_f) &
                                                          +coefs%coef_htfrtc%coef_l(t_opt_ly_i(i)+1,p_opt_ly_i(i),  1,i_f)) &
                                    +p_opt_ly_r(i)*(-coefs%coef_htfrtc%coef_l(t_opt_ly_i(i),  p_opt_ly_i(i)+1,  1,i_f) &
                                                    +coefs%coef_htfrtc%coef_l(t_opt_ly_i(i)+1,p_opt_ly_i(i)+1,  1,i_f))) &
                          +sg_ly*((1.0-p_opt_ly_r(i))*(-coefs%coef_htfrtc%coef_l(t_opt_ly_i(i),  p_opt_ly_i(i),  2,i_f) &
                                                       +coefs%coef_htfrtc%coef_l(t_opt_ly_i(i)+1,p_opt_ly_i(i),  2,i_f)) &
                                 +p_opt_ly_r(i)*(-coefs%coef_htfrtc%coef_l(t_opt_ly_i(i),  p_opt_ly_i(i)+1,  2,i_f) &
                                                 +coefs%coef_htfrtc%coef_l(t_opt_ly_i(i)+1,p_opt_ly_i(i)+1,  2,i_f))))
               ENDFORALL

               FORALL(i=1:nlayers)
                  dtrdq(i)=-tr(i)*mpath(i)*mft(i)* &
                          (dq(i) &
                          +q_ly(i)/(coefs%coef_htfrtc%val_mf_nl(q_opt_ly_i(i)+1,1)-coefs%coef_htfrtc%val_mf_nl(q_opt_ly_i(i),1)) &
                          *(pt(i,1)*(coefs%coef_htfrtc%coef_nl(qt(i,1)+1, p_opt_ly_i(i)  , 1,i_f) &
                                    -coefs%coef_htfrtc%coef_nl(qt(i,1)  , p_opt_ly_i(i)  , 1,i_f)) &
                           +pt(i,2)*(coefs%coef_htfrtc%coef_nl(qt(i,2)+1, p_opt_ly_i(i)  , 1,i_f) &
                                    -coefs%coef_htfrtc%coef_nl(qt(i,2)  , p_opt_ly_i(i)  , 1,i_f)) &
                           +pt(i,3)*(coefs%coef_htfrtc%coef_nl(qt(i,1)+1, p_opt_ly_i(i)+1, 1,i_f) &
                                    -coefs%coef_htfrtc%coef_nl(qt(i,1)  , p_opt_ly_i(i)+1, 1,i_f)) &
                           +pt(i,4)*(coefs%coef_htfrtc%coef_nl(qt(i,2)+1, p_opt_ly_i(i)+1, 1,i_f) &
                                    -coefs%coef_htfrtc%coef_nl(qt(i,2)  , p_opt_ly_i(i)+1, 1,i_f))))
               ENDFORALL

               FORALL(i=1:nlayers)
                  dtrdo3(i)=-tr(i)*mpath(i)*mft(i)*do3(i)
               ENDFORALL

             ELSE IF (coefs%coef_htfrtc%opt_prop_type==2) THEN

               FORALL(i=1:nlayers)
                  dtrdt(i)=-tr(i)*mpath(i)*mft(i)* &
                          (q_ly(i)*((1.0-p_opt_ly_r(i)) *(coefs%coef_htfrtc%coef_nl(2,p_opt_ly_i(i)  ,1,i_f)  &
                                     +two*t_ly(i)*        coefs%coef_htfrtc%coef_nl(3,p_opt_ly_i(i)  ,1,i_f)  &
                                     +            q_ly(i)*coefs%coef_htfrtc%coef_nl(5,p_opt_ly_i(i)  ,1,i_f)  &
                                     +two*t_ly(i)*q_ly(i)*coefs%coef_htfrtc%coef_nl(6,p_opt_ly_i(i)  ,1,i_f)) &
                                  +p_opt_ly_r(i)        *(coefs%coef_htfrtc%coef_nl(2,p_opt_ly_i(i)+1,1,i_f)  &
                                     +two*t_ly(i)*        coefs%coef_htfrtc%coef_nl(3,p_opt_ly_i(i)+1,1,i_f)  &
                                     +            q_ly(i)*coefs%coef_htfrtc%coef_nl(5,p_opt_ly_i(i)+1,1,i_f)  &
                                     +two*t_ly(i)*q_ly(i)*coefs%coef_htfrtc%coef_nl(6,p_opt_ly_i(i)+1,1,i_f))) &
                           +o3_ly(i)*((1.0-p_opt_ly_r(i)) *(coefs%coef_htfrtc%coef_l(2,p_opt_ly_i(i)  ,1,i_f)  &
                                                +two*t_ly(i)*coefs%coef_htfrtc%coef_l(3,p_opt_ly_i(i)  ,1,i_f)) &
                                   +p_opt_ly_r(i)      *(    coefs%coef_htfrtc%coef_l(2,p_opt_ly_i(i)+1,1,i_f)  &
                                                +two*t_ly(i)*coefs%coef_htfrtc%coef_l(3,p_opt_ly_i(i)+1,1,i_f))) &
                           +sg_ly*((1.0-p_opt_ly_r(i)) *(    coefs%coef_htfrtc%coef_l(2,p_opt_ly_i(i)  ,2,i_f)  &
                                                +two*t_ly(i)*coefs%coef_htfrtc%coef_l(3,p_opt_ly_i(i)  ,2,i_f)) &
                                   +p_opt_ly_r(i)      *(    coefs%coef_htfrtc%coef_l(2,p_opt_ly_i(i)+1,2,i_f)  &
                                                +two*t_ly(i)*coefs%coef_htfrtc%coef_l(3,p_opt_ly_i(i)+1,2,i_f))))
                ENDFORALL

                FORALL(i=1:nlayers)
                   dtrdq(i)=-tr(i)*mpath(i)*mft(i)* &
                           (dq(i)+ &
                           q_ly(i)*((1.0-p_opt_ly_r(i))*(coefs%coef_htfrtc%coef_nl(4,p_opt_ly_i(i)  ,1,i_f) &
                                              +t_ly(i) * coefs%coef_htfrtc%coef_nl(5,p_opt_ly_i(i)  ,1,i_f) &
                                              +t2_ly(i)* coefs%coef_htfrtc%coef_nl(6,p_opt_ly_i(i)  ,1,i_f)) &
                                  +p_opt_ly_r(i)       *(coefs%coef_htfrtc%coef_nl(4,p_opt_ly_i(i)+1,1,i_f) &
                                              +t_ly(i) * coefs%coef_htfrtc%coef_nl(5,p_opt_ly_i(i)+1,1,i_f) &
                                              +t2_ly(i)* coefs%coef_htfrtc%coef_nl(6,p_opt_ly_i(i)+1,1,i_f))))
                ENDFORALL

               FORALL(i=1:nlayers)
                  dtrdo3(i)=-tr(i)*mpath(i)*mft(i)*do3(i)
               ENDFORALL

             ENDIF

             FORALL(i=1:nlevels)
                dbdt(i)=t_b_inc_m1*(coefs%coef_htfrtc%coef_b(t_b_i(i)+1,i_f)-coefs%coef_htfrtc%coef_b(t_b_i(i),i_f))
             ENDFORALL
             db_surfdt=t_b_inc_m1*(coefs%coef_htfrtc%coef_b(t_b_surf_i+1,i_f)-coefs%coef_htfrtc%coef_b(t_b_surf_i,i_f))

             FORALL(i=1:nlayers)
                dblydowndt(i)=dbdt(i)+lt(i)*(dbdt(i+1)-dbdt(i))
                dblyupdt(i)=dbdt(i+1)+lt(i)*(dbdt(i)-dbdt(i+1))
             ENDFORALL

             tot=product(tr(1:surf_level))
             DO i=1,surf_level
                totdown(i)  =product(tr(i+1:surf_level))*tot*(unity-surf_em(i_f))
                drdowndt(i) =(dblydowndt(i)*(unity-tr(i))+(rdown(i)-blydown(i))*dtrdt(i))*totdown(i)
                drdowndq(i) =(rdown(i)-blydown(i))*dtrdq(i)*totdown(i)
                drdowndo3(i)=(rdown(i)-blydown(i))*dtrdo3(i)*totdown(i)
                totup(i)    =product(tr(1:i-1))
                drupdt(i)   =(dblyupdt(i)*(unity-tr(i))+(rup(i+1)-blyup(i))*dtrdt(i))*totup(i)
                drupdq(i)   =(rup(i+1)-blyup(i))*dtrdq(i)*totup(i)
                drupdo3(i)  =(rup(i+1)-blyup(i))*dtrdo3(i)*totup(i)
             ENDDO

             drcdt(i_f,1)=ahalf*(drdowndt(1)+drupdt(1))*normm1(i_f)
             drcdq(i_f,1)=ahalf*(drdowndq(1)+drupdq(1))*normm1(i_f)
             drcdo3(i_f,1)=ahalf*(drdowndo3(1)+drupdo3(1))*normm1(i_f)
             DO i=2, surf_level
                drcdt(i_f,i)=ahalf*(drdowndt(i-1)+drdowndt(i)+drupdt(i-1)+drupdt(i))*normm1(i_f)
                drcdq(i_f,i)=ahalf*(drdowndq(i-1)+drdowndq(i)+drupdq(i-1)+drupdq(i))*normm1(i_f)
                drcdo3(i_f,i)=ahalf*(drdowndo3(i-1)+drdowndo3(i)+drupdo3(i-1)+drupdo3(i))*normm1(i_f)
             ENDDO
             drcdt(i_f,surf_level+1)=ahalf*(drdowndt(surf_level)+drupdt(surf_level))*normm1(i_f)
             drcdq(i_f,surf_level+1)=ahalf*(drdowndq(surf_level)+drupdq(surf_level))*normm1(i_f)
             drcdo3(i_f,surf_level+1)=ahalf*(drdowndo3(surf_level)+drupdo3(surf_level))*normm1(i_f)
             DO i=surf_level+2, nlevels
             drcdt(i_f,i)=zero
             drcdq(i_f,i)=zero
             drcdo3(i_f,i)=zero
             ENDDO

             drcdst(i_f)=db_surfdt*surf_em(i_f)*tot*normm1(i_f)
             drcds2mt(i_f)=drcdt(i_f,surf_level+1)
             drcds2mq(i_f)=drcdq(i_f,surf_level+1)
             drcds2mp(i_f)=((rdown(surf_level)-blydown(surf_level))*(unity-surf_em(i_f))+(rup(surf_level+1)-blyup(surf_level))) &
                          *tr(surf_level)*(-tausl(surf_level))/dp(surf_level)*totup(surf_level)*normm1(i_f)
             drcds2mu(i_f)=b_surf*dsurf_emdu(i_f)*tot*normm1(i_f)
             drcds2mv(i_f)=b_surf*dsurf_emdv(i_f)*tot*normm1(i_f)

             ENDIF ! do_k

          ENDDO ! n_f

       !PC scores
       FORALL(i_pc=1:n_pc)
          pccomp%pcscores((i_p-1)*n_pc+i_pc)=sum(rc(1:n_f)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
       END FORALL ! n_pc

       IF(do_k) THEN

          FORALL(i_pc=1:n_pc)
             FORALL(i=1:nlayers)
                profiles_k_pc((i_p-1)*n_pc+i_pc)%t(i)=sum(drcdt(1:n_f,i)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
                profiles_k_pc((i_p-1)*n_pc+i_pc)%q(i)=sum(drcdq(1:n_f,i)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
                profiles_k_pc((i_p-1)*n_pc+i_pc)%o3(i)=sum(drcdo3(1:n_f,i)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
             ENDFORALL
                profiles_k_pc((i_p-1)*n_pc+i_pc)%skin%t=sum(drcdst(1:n_f)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
                profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%t=sum(drcds2mt(1:n_f)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
                profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%q=sum(drcds2mq(1:n_f)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
                profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%p=sum(drcds2mp(1:n_f)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
                profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%u=sum(drcds2mu(1:n_f)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
                profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%v=sum(drcds2mv(1:n_f)*coefs%coef_htfrtc%coef_pdt(1:n_f,i_pc))
          END FORALL ! n_pc

       ENDIF

       IF(opts%htfrtc_opts%reconstruct) THEN

         !Radiances
         FORALL(i_ch=1:n_ch)
            pccomp%total_pccomp((i_p-1)*n_ch+i_ch)=coefs%coef_htfrtc%ch_mean(i_ch)+&
              sum(pccomp%pcscores((i_p-1)*n_pc+1:i_p*n_pc)*coefs%coef_htfrtc%pc(1:n_pc,i_ch))
         END FORALL

         !Brightness temperatures
         FORALL(i_ch=1:n_ch)
            pccomp%bt_pccomp((i_p-1)*n_ch+i_ch)=planck_c2_wn(i_ch)/ &
                log(unity+planck_c1_wn3(i_ch)/pccomp%total_pccomp((i_p-1)*n_ch+i_ch))
         ENDFORALL

         IF(do_k) THEN

!          FORALL(i_ch=1:n_ch)
         DO i_ch = 1, n_ch
            profiles_k_rec((i_p-1)*n_ch+i_ch)%t(:) = 0._jprb
            profiles_k_rec((i_p-1)*n_ch+i_ch)%q(:) = 0._jprb
            profiles_k_rec((i_p-1)*n_ch+i_ch)%o3(:) = 0._jprb
            profiles_k_rec((i_p-1)*n_ch+i_ch)%skin%t = 0._jprb
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%t = 0._jprb
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%q = 0._jprb
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%p = 0._jprb
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%u = 0._jprb
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%v = 0._jprb
!             FORALL(i_pc=1:n_pc)
            DO i_pc = 1, n_pc
!                 FORALL(i=1:nlayers)
                DO i = 1, nlayers
                  profiles_k_rec((i_p-1)*n_ch+i_ch)%t(i)=profiles_k_rec((i_p-1)*n_ch+i_ch)%t(i) + &
                    profiles_k_pc((i_p-1)*n_pc+i_pc)%t(i)*coefs%coef_htfrtc%pc(i_pc,i_ch)
                  profiles_k_rec((i_p-1)*n_ch+i_ch)%q(i)=profiles_k_rec((i_p-1)*n_ch+i_ch)%q(i) + &
                    profiles_k_pc((i_p-1)*n_pc+i_pc)%q(i)*coefs%coef_htfrtc%pc(i_pc,i_ch)
                  profiles_k_rec((i_p-1)*n_ch+i_ch)%o3(i)=profiles_k_rec((i_p-1)*n_ch+i_ch)%o3(i) + &
                    profiles_k_pc((i_p-1)*n_pc+i_pc)%o3(i)*coefs%coef_htfrtc%pc(i_pc,i_ch)
!                ENDFORALL
               ENDDO
               profiles_k_rec((i_p-1)*n_ch+i_ch)%skin%t=profiles_k_rec((i_p-1)*n_ch+i_ch)%skin%t + &
                 profiles_k_pc((i_p-1)*n_pc+i_pc)%skin%t*coefs%coef_htfrtc%pc(i_pc,i_ch)
               profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%t=profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%t + &
                 profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%t*coefs%coef_htfrtc%pc(i_pc,i_ch)
               profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%q=profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%q + &
                 profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%q*coefs%coef_htfrtc%pc(i_pc,i_ch)
               profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%p=profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%p + &
                 profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%p*coefs%coef_htfrtc%pc(i_pc,i_ch)
               profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%u=profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%u + &
                 profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%u*coefs%coef_htfrtc%pc(i_pc,i_ch)
               profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%v=profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%v + &
                 profiles_k_pc((i_p-1)*n_pc+i_pc)%s2m%v*coefs%coef_htfrtc%pc(i_pc,i_ch)
!             ENDFORALL
            ENDDO
!          ENDFORALL
         ENDDO

         IF(opts%rt_all%switchrad) THEN
         DO i_ch = 1, n_ch
            swrad_tmp(i_ch)=planck_c1_wn3(i_ch)*(pccomp%bt_pccomp((i_p-1)*n_ch+i_ch))**2 &
            /(planck_c2_wn(i_ch)*pccomp%total_pccomp((i_p-1)*n_ch+i_ch)*&
            (pccomp%total_pccomp((i_p-1)*n_ch+i_ch)+planck_c1_wn3(i_ch)))
         ENDDO
         DO i_ch = 1, n_ch
            DO i = 1, nlayers
               profiles_k_rec((i_p-1)*n_ch+i_ch)%t(i)=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%t(i)
               profiles_k_rec((i_p-1)*n_ch+i_ch)%q(i)=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%q(i)
               profiles_k_rec((i_p-1)*n_ch+i_ch)%o3(i)=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%o3(i)
            ENDDO
            profiles_k_rec((i_p-1)*n_ch+i_ch)%skin%t=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%skin%t
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%t=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%t
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%q=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%q
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%p=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%p
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%u=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%u
            profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%v=swrad_tmp(i_ch)*profiles_k_rec((i_p-1)*n_ch+i_ch)%s2m%v
         ENDDO
         ENDIF

         ENDIF

       ENDIF ! reconstruct

    ENDDO ! i_p
    CATCH
  END SUBROUTINE htfrtc_interface

END MODULE rttov_htfrtc_interface_mod
